/*
 *  This file is part of nzbget. See <https://nzbget.com>.
 *
 *  Copyright (C) 2023-2025 Denis <denis@nzbget.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef XML_H
#define XML_H

#include <iostream>
#include <libxml/tree.h>

namespace Xml
{
using XmlNodePtr = xmlNodePtr;
std::string Serialize(const xmlNodePtr rootNode);
void AddNewNode(xmlNodePtr rootNode, const char* name, const char* type, const char* value);
xmlNodePtr CreateStructNode();
void AddArrayNode(xmlNodePtr structNode, const char* name, std::vector<xmlNodePtr> children);
const char* BoolToStr(bool value) noexcept;
}  // namespace Xml

#endif
