#ifndef __CRC_ZEROPAD_H
#define __CRC_ZEROPAD_H

#include <par2/osinfo/stdint.h>

#ifdef __cplusplus
extern "C"
#endif
uint32_t crc_zeroPad(uint32_t crc, uint64_t zeroPad);

#endif /* __CRC_ZEROPAD_H */
